# This file is a part of Remoulade.
#
# Copyright (C) 2017,2018 CLEARTYPE SRL <bogdan@cleartype.io>
#
# Remoulade is free software; you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or (at
# your option) any later version.
#
# Remoulade is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public
# License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
import time
from typing import Any, Dict, Set

from ..backend import ForgottenResult, Missing, ResultBackend


class StubBackend(ResultBackend):
    """An in-memory result backend.  For use in unit tests.

    Parameters:
      namespace(str): A string with which to prefix result keys.
      encoder(Encoder): The encoder to use when storing and retrieving
        result data.  Defaults to :class:`.JSONEncoder`.
    """

    results: Dict[str, Any] = {}
    group_completions: Dict[str, Set[str]] = {}

    def _get(self, message_key: str, forget: bool = False):
        if forget:
            data, expiration = self.results.get(message_key, (None, None))
            if data is not None:
                self.results[message_key] = self.encoder.encode(ForgottenResult.asdict()), expiration
        else:
            data, expiration = self.results.get(message_key, (None, None))

        if data is not None and time.monotonic() < expiration:
            return self.encoder.decode(data)
        return Missing

    def _store(self, message_keys, results, ttl):
        for (message_key, result) in zip(message_keys, results):
            result_data = self.encoder.encode(result)
            expiration = time.monotonic() + int(ttl / 1000)
            self.results[message_key] = (result_data, expiration)

    def _delete(self, key: str):
        del self.results[key]

    def increment_group_completion(self, group_id: str, message_id: str) -> int:
        group_completion_key = self.build_group_completion_key(group_id)
        completed = self.results.get(group_completion_key, set()) | {message_id}
        self.results[group_completion_key] = completed
        return len(completed)



import argparse
import importlib
import logging
import signal
import sys

from remoulade import get_logger, get_scheduler

logformat = "[%(asctime)s] [PID %(process)d] [%(threadName)s] [%(name)s] [%(levelname)s] %(message)s"


def parse_arguments():
    parser = argparse.ArgumentParser(
        prog="remoulade-scheduler",
        description="Run remoulade scheduler",
        formatter_class=argparse.RawDescriptionHelpFormatter,
    )
    parser.add_argument("modules", metavar="module", nargs="*", help="additional python modules to import")
    parser.add_argument("--path", "-P", default=".", nargs="*", type=str, help="the module import path (default: .)")
    parser.add_argument("--verbose", "-v", action="count", default=0, help="turn on verbose log output")
    return parser.parse_args()


def main():

    args = parse_arguments()

    for path in args.path:
        sys.path.insert(0, path)

    for module in args.modules:
        importlib.import_module(module)

    logging.basicConfig(level=logging.INFO if not args.verbose else logging.DEBUG, format=logformat)
    logger = get_logger("remoulade", "Scheduler")

    def signal_handler(signal, frame):
        logger.debug("Remoulade scheduler is shutting down")
        sys.exit(0)

    signal.signal(signal.SIGINT, signal_handler)
    signal.signal(signal.SIGTERM, signal_handler)

    logger.debug("Remoulade scheduler start")
    sys.exit(get_scheduler().start())



""" This file describe the API to get the state of messages """
import datetime
import sys
from collections import defaultdict
from operator import itemgetter
from typing import Dict, Iterable, List

from flask import Flask, request
from marshmallow import ValidationError
from typing_extensions import DefaultDict, TypedDict
from werkzeug.exceptions import HTTPException, NotFound

import remoulade
from remoulade import get_broker, get_scheduler
from remoulade.errors import NoResultBackend, NoScheduler, RemouladeError
from remoulade.result import Result
from remoulade.results import ResultMissing

from .schema import MessageSchema, PageSchema

app = Flask(__name__)


def sort_dicts(data, column, reverse=False):
    """Sort an array of dicts by a given column"""
    data_none = [item for item in data if item.get(column) is None]
    data = sorted((item for item in data if item.get(column) is not None), key=itemgetter(column), reverse=reverse)
    data.extend(data_none)
    return data


def dict_has(item, keys, value):
    """Check if the value of some key in keys has a value"""
    return chr(0).join([str(item[k]) for k in keys if item.get(k)]).lower().find(value) >= 0


class GroupMessagesT(TypedDict):
    group_id: str
    messages: List[dict]


@app.route("/messages/states")
def get_states():
    args = PageSchema().load(request.args.to_dict())
    backend = remoulade.get_broker().get_state_backend()
    data = [s.as_dict() for s in backend.get_states()]
    if args.get("search_value"):
        keys = ["message_id", "name", "actor_name", "args", "kwargs", "options"]
        value = args["search_value"].lower()
        data = [item for item in data if dict_has(item, keys, value)]

    if args.get("sort_column"):
        reverse = args.get("sort_direction") == "desc"
        sort_column = args["sort_column"]
        data = sort_dicts(data, sort_column, reverse)

    return {"data": data[args["offset"] : args["size"] + args["offset"]], "count": len(data)}


@app.route("/messages/state/<message_id>")
def get_state(message_id):
    backend = remoulade.get_broker().get_state_backend()
    data = backend.get_state(message_id)
    if data is None:
        raise NotFound("message_id = {} does not exist".format(message_id))
    return data.as_dict(encode_args=True)


@app.route("/messages/cancel/<message_id>", methods=["POST"])
def cancel_message(message_id):
    backend = remoulade.get_broker().get_cancel_backend()
    backend.cancel([message_id])
    return {"result": "ok"}


@app.route("/messages/requeue/<message_id>")
def requeue_message(message_id):
    broker = remoulade.get_broker()
    backend = broker.get_state_backend()
    state = backend.get_state(message_id)
    actor = broker.get_actor(state.actor_name)
    payload = {"args": state.args, "kwargs": state.kwargs}
    pipe_target = state.options.get("pipe_target")
    if pipe_target is None:
        actor.send_with_options(**payload, **state.options)
        return {"result": "ok"}
    else:
        return {"error": "requeue message in a pipeline not supported"}, 400


@app.route("/messages/result/<message_id>")
def get_results(message_id):
    from ..message import get_encoder

    max_size = 1e4
    try:
        result = Result(message_id=message_id).get()
        encoded_result = get_encoder().encode(result).decode("utf-8")
        size_result = sys.getsizeof(encoded_result)
        if size_result >= max_size:
            encoded_result = "The result is too big {}M".format(size_result / 1e6)
        return {"result": encoded_result}
    except ResultMissing:
        return {"result": "result is missing"}
    except NoResultBackend:
        return {"result": "no result backend"}
    except (UnicodeDecodeError, TypeError):
        return {"result": "non serializable result"}


@app.route("/scheduled/jobs")
def get_scheduled_jobs():
    try:
        scheduler = get_scheduler()
    except NoScheduler:
        return {"result": []}
    scheduled_jobs = scheduler.get_redis_schedule()
    return {"result": [job.as_dict() for job in scheduled_jobs.values()]}


@app.route("/messages", methods=["POST"])
def enqueue_message():
    payload = MessageSchema().load(request.json)
    actor = get_broker().get_actor(payload.pop("actor_name"))
    options = payload.pop("options") or {}
    actor.send_with_options(**payload, **options)
    return {"result": "ok"}


@app.route("/actors")
def get_actors():
    return {"result": [actor.as_dict() for actor in get_broker().actors.values()]}


@app.route("/groups")
def get_groups():
    args = PageSchema().load(request.args.to_dict())
    backend = remoulade.get_broker().get_state_backend()
    groups_by_id: DefaultDict[str, List[Dict]] = defaultdict(list)
    states = (state for state in backend.get_states() if state.group_id)

    if args.get("search_value"):
        keys = ["message_id", "name", "actor_name", "group_id"]
        value = args["search_value"].lower()
        states = [state for state in states if dict_has(state.as_dict(), keys, value)]  # type: ignore

    for state in states:
        groups_by_id[state.group_id].append(state.as_dict(exclude_keys=("args", "kwargs", "options")))

    groups: Iterable[GroupMessagesT] = (
        {"group_id": group_id, "messages": messages} for group_id, messages in groups_by_id.items()
    )
    sorted_groups: List[GroupMessagesT] = sorted(
        groups, key=lambda x: x["messages"][0].get("enqueued_datetime") or datetime.datetime.min, reverse=True
    )
    return {"data": sorted_groups[args["offset"] : args["size"] + args["offset"]], "count": len(sorted_groups)}


@app.route("/options")
def get_options():
    broker = remoulade.get_broker()
    return {"options": list(broker.actor_options)}


@app.errorhandler(RemouladeError)
def remoulade_exception(e):
    return {"error": str(e)}, 500


@app.errorhandler(HTTPException)
def http_exception(e):
    return {"error": str(e)}, e.code


@app.errorhandler(ValidationError)
def validation_error(e):
    return {"error": e.normalized_messages()}, 400



# This file is a part of Remoulade.
#
# Copyright (C) 2017,2018 CLEARTYPE SRL <bogdan@cleartype.io>
#
# Remoulade is free software; you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or (at
# your option) any later version.
#
# Remoulade is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public
# License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import warnings

from .stub import StubBackend

try:
    from .redis import RedisBackend
except ImportError:  # pragma: no cover
    warnings.warn(
        "RedisBackend is not available.  Run `pip install remoulade[redis]` " "to add support for that backend.",
        ImportWarning,
    )


__all__ = ["StubBackend", "RedisBackend"]



# SPDX-License-Identifier: LPGL-3.0
# Copyright (C) 2019 CLEARTYPE SRL <bogdan@cleartype.io>
# Based in current_message.py from the dramatiq project

from threading import local
from typing import TYPE_CHECKING

from .middleware import Middleware

if TYPE_CHECKING:
    from .. import Broker, Message


class CurrentMessage(Middleware):

    local_data = local()

    @classmethod
    def get_current_message(cls) -> "Message":
        """Get the message that triggered the current actor.  Messages
        are thread local so this returns ``None`` when called outside
        of actor code.
        """
        return getattr(cls.local_data, "current_message", None)

    def before_process_message(self, broker: "Broker", message: "Message") -> None:
        self.local_data.current_message = message

    def after_process_message(self, broker: "Broker", message: "Message", *, result=None, exception=None) -> None:
        self.local_data.current_message = None



import argparse
import importlib
import sys

from remoulade import get_broker


def parse_arguments():
    parser = argparse.ArgumentParser(
        prog="remoulade-ls",
        description="List of remoulade actors",
        formatter_class=argparse.RawDescriptionHelpFormatter,
    )
    parser.add_argument("modules", metavar="module", nargs="*", help="additional python modules to import")
    parser.add_argument("--path", "-P", default=".", nargs="*", type=str, help="the module import path (default: .)")
    return parser.parse_args()


def main():

    args = parse_arguments()

    for path in args.path:
        sys.path.insert(0, path)

    for module in args.modules:
        importlib.import_module(module)

    print("List of available actors:")
    for actor in get_broker().actors:
        print(actor)



try:
    from .redis import RedisBackend
    from .stub import StubBackend
except ImportError:  # pragma: no cover
    import warnings

    warnings.warn(
        "RedisBackend is not available.  Run `pip install remoulade[redis]` " "to add support for that backend.",
        ImportWarning,
    )

__all__ = ["RedisBackend", "StubBackend"]



import sys
from collections import namedtuple
from enum import Enum
from typing import Dict, List

from dateutil.parser import parse

from ..encoder import Encoder
from .errors import InvalidStateError


class StateNamesEnum(Enum):
    """Contains the possible states that can have a message"""

    Started = "Started"  # a `Message` that has not been processed
    Pending = "Pending"  # a `Message` that has been enqueued
    Skipped = "Skipped"  # a `Message` that has been skipped
    Canceled = "Canceled"  # a `Message` that has been canceled
    Failure = "Failure"  # a `Message` that has been processed and raised an Exception
    Success = "Success"  # a `Message` that has been processed and does not raise an Exception


#: A type alias representing states in the database
class State(
    namedtuple(
        "State",
        (
            "message_id",
            "name",
            "actor_name",
            "args",
            "kwargs",
            "options",
            "priority",
            "progress",
            "enqueued_datetime",
            "started_datetime",
            "end_datetime",
            "group_id",
        ),
    )
):
    """Catalog Class, it storages the state
    Parameters:
        name: Name of the state
        args: List of arguments in the state(name)
    """

    def __new__(
        cls,
        message_id,
        name=None,
        *,
        actor_name=None,
        args=None,
        kwargs=None,
        options=None,
        priority=None,
        progress=None,
        enqueued_datetime=None,
        started_datetime=None,
        end_datetime=None,
        group_id=None,
    ):

        if name and name not in list(StateNamesEnum):
            raise InvalidStateError("The {} State is not defined".format(name))
        return super().__new__(
            cls,
            message_id,
            name,
            actor_name,
            args,
            kwargs,
            options,
            priority,
            progress,
            enqueued_datetime,
            started_datetime,
            end_datetime,
            group_id,
        )

    def as_dict(self, exclude_keys=(), encode_args=False):
        """Transform a State into a dict, can exclude some keys"""
        as_dict = {
            key: value for (key, value) in self._asdict().items() if value is not None and key not in exclude_keys
        }
        datetime_keys = ["enqueued_datetime", "started_datetime", "end_datetime"]
        for key in datetime_keys:
            if key in as_dict:
                as_dict[key] = as_dict[key].isoformat()
        if self.name:
            as_dict["name"] = self.name.value
        if encode_args:
            from ..message import get_encoder

            for key in (item for item in ["args", "kwargs", "options"] if item in as_dict):
                try:
                    as_dict[key] = get_encoder().encode(as_dict[key]).decode("utf-8")
                except (UnicodeDecodeError, TypeError):
                    as_dict[key] = "encoded_data"
        return as_dict

    @classmethod
    def from_dict(cls, input_dict: Dict) -> "State":
        if "name" in input_dict:
            input_dict["name"] = StateNamesEnum(input_dict["name"])
        datetime_keys = ["enqueued_datetime", "started_datetime", "end_datetime"]
        for key in datetime_keys:
            if key in input_dict:
                input_dict[key] = parse(input_dict[key])
        return cls(**input_dict)


class StateBackend:
    """ABC for  state backends.

    Parameters:
      namespace(str): The logical namespace under which the data
        should be stored.
      encoder(Encoder): The encoder to use when storing and retrieving
        result data.  Defaults to :class:`.JSONEncoder`.
      max_size(int): Maximum size of arguments allow to storage
        in the database, default 2MB
    """

    namespace = "remoulade-state*"

    def __init__(self, *, namespace: str = "remoulade-state", encoder: Encoder = None, max_size=2e6):
        from ..message import get_encoder

        self.namespace = namespace
        self.encoder = encoder or get_encoder()
        self.max_size = max_size

    def _build_message_key(self, message_id: str) -> str:  # noqa: F821
        """Given a message id, return its globally-unique key.

        Parameters:
          message_id(str)

        Returns:
          str
        """
        return f"{self.namespace}:{message_id}"

    def get_state(self, message_id: str) -> State:
        """Get the state with a message_id from the backend.

        Parameters:
            message_id(str)

        """
        raise NotImplementedError(f"{type(self).__name__} does not implement get_state")

    def set_state(self, state: State, ttl: int = 3600) -> None:
        """Save a message in the backend if it does not exist,
            otherwise update it.

        Parameters:
            state(State)
            ttl(seconds): The time to keep that state in the backend
             default is one hour(3600 seconds)
        """
        raise NotImplementedError(f"{type(self).__name__} does not implement set_state")

    def get_states(self) -> List[State]:
        """Return all the states in the backend"""
        raise NotImplementedError(f"{type(self).__name__} does not implement get_all_messages")

    def _encode_dict(self, data):
        """Return the (keys, values) of a dictionary encoded"""
        encoded_data = {}
        for (key, value) in data.items():
            encoded_value = self.encoder.encode(value)
            if sys.getsizeof(encoded_value) <= self.max_size:
                encoded_data[self.encoder.encode(key)] = self.encoder.encode(value)
        return encoded_data

    def _decode_dict(self, data):
        """Return the (keys, values) of a dictionary decoded"""
        decoded_data = {}
        for (key, value) in data.items():
            decoded_data[self.encoder.decode(key)] = self.encoder.decode(value)
        return decoded_data



from inspect import Parameter, signature


def get_actor_arguments(actor):
    params = signature(actor.fn).parameters

    def parse_type(rawtype):
        if str(rawtype).startswith("typing"):
            return str(rawtype)
        try:
            return str(rawtype.__name__)
        except AttributeError:
            return str(rawtype)

    args = []
    for param in params.values():
        arg = {"name": param.name}
        if param.annotation != Parameter.empty:
            arg["type"] = parse_type(param.annotation)
        if param.default != Parameter.empty:
            if param.default is None:
                arg["default"] = ""
            else:
                arg["default"] = str(param.default)

        args.append(arg)

    return args



# This file is a part of Remoulade.
#
# Copyright (C) 2017,2018 CLEARTYPE SRL <bogdan@cleartype.io>
#
# Remoulade is free software; you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or (at
# your option) any later version.
#
# Remoulade is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public
# License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import time
from collections import namedtuple
from contextlib import contextmanager
from typing import Any, Dict, Iterable, List, Type, Union

from ..encoder import Encoder
from ..helpers import compute_backoff
from .errors import ErrorStored, MessageIdsMissing, ResultMissing, ResultTimeout

#: The default timeout for blocking get operations in milliseconds.
DEFAULT_TIMEOUT = 10000

#: The minimum amount of time in ms to wait between polls.
BACKOFF_FACTOR = 100


#: Canary value that is returned when a result hasn't been set yet.
class Missing:
    pass


#: A type alias representing backend results.
class BackendResult(namedtuple("BackendResult", ("result", "error", "forgot"))):
    def __new__(cls, *, result, error, forgot=False):
        return super().__new__(cls, result, error, forgot)

    def asdict(self):
        return self._asdict()


ForgottenResult = BackendResult(result=None, error=None, forgot=True)


class ResultBackend:
    """ABC for result backends.

    Parameters:
      namespace(str): The logical namespace under which the data
        should be stored.
      encoder(Encoder): The encoder to use when storing and retrieving
        result data.  Defaults to :class:`.JSONEncoder`.
    """

    def __init__(self, *, namespace: str = "remoulade-results", encoder: Encoder = None):
        from ..message import get_encoder

        self.namespace = namespace
        self.encoder = encoder or get_encoder()

    def get_result(
        self,
        message_id: str,
        *,
        block: bool = False,
        timeout: int = None,
        forget: bool = False,
        raise_on_error: bool = True,
    ) -> BackendResult:
        """Get a result from the backend.

        Parameters:
          message_id(str)
          block(bool): Whether or not to block until a result is set.
          timeout(int): The maximum amount of time, in ms, to wait for
            a result when block is True.  Defaults to 10 seconds.
          forget(bool): Whether or not the result need to be kept.
          raise_on_error(bool): raise an error if the result stored in
            an error

        Raises:
          ResultMissing: When block is False and the result isn't set.
          ResultTimeout: When waiting for a result times out.
          ErrorStored: When the result is an error and raise_on_error is True

        Returns:
          object: The result.
        """
        if timeout is None:
            timeout = DEFAULT_TIMEOUT

        end_time = time.monotonic() + timeout / 1000
        message_key = self.build_message_key(message_id)

        attempts = 0
        while True:
            result = self._get(message_key, forget)
            if result is Missing and block:
                attempts, delay = compute_backoff(attempts, min_backoff=BACKOFF_FACTOR)
                delay /= 1000
                if time.monotonic() + delay > end_time:
                    raise ResultTimeout(message_id)

                time.sleep(delay)
                continue

            elif result is Missing:
                raise ResultMissing(message_id)

            else:
                backend_result = BackendResult(**result)  # type: ignore
                return self.process_result(backend_result, raise_on_error)

    @staticmethod
    def process_result(result: BackendResult, raise_on_error: bool):
        """Raise an error if the result is an error and raise_on_error is true

        Parameters:
          result(Result): the result retrieved from the backend
          raise_on_error(boolean): raise an error if the result stored in an error
        """
        if result.error:
            error = ErrorStored(result.error)
            if raise_on_error:
                raise error
            return error
        return result.result

    def increment_group_completion(self, group_id: str, message_id: str) -> int:
        raise NotImplementedError(f"{type(self).__name__} does not implement increment_group_completion()")

    def store_result(self, message_id: str, result: BackendResult, ttl: int) -> None:
        """Store a result in the backend.

        Parameters:
          message_id(str)
          result(BackendResult): Must be serializable.
          ttl(int): The maximum amount of time the result may be stored in the backend for.
        """
        return self.store_results([message_id], [result], ttl)

    def store_results(self, message_ids: Iterable[str], results: Iterable[BackendResult], ttl: int) -> None:
        """Store multiple results in the backend.

        Parameters:
          message_ids(Iterable[str])
          results(BackendResult): Must be serializable.
          ttl(int): The maximum amount of time the result may be stored in the backend for.
        """
        message_keys = [self.build_message_key(message_id) for message_id in message_ids]
        return self._store(message_keys, [r._asdict() for r in results], ttl)

    def forget_results(self, message_ids: List[str], ttl: int):
        """Forget the results associated to the given message_id"""
        result = ForgottenResult.asdict()
        message_keys = [self.build_message_key(message_id) for message_id in message_ids]
        self._store(message_keys, [result] * len(message_keys), ttl)

    def build_message_key(self, message_id: str) -> str:
        """Given a message, return its globally-unique key"""
        return f"{self.namespace}:{message_id}"

    def get_status(self, message_ids: Iterable[str]) -> int:
        """Given a list of messages ids return the number of messages with a result stored"""
        count = 0
        for message_id in message_ids:
            message_key = self.build_message_key(message_id)
            result = self._get(message_key, forget=False)
            if result != Missing:
                count += 1
        return count

    @staticmethod
    def build_group_message_id_key(group_id: str) -> str:
        return f"remoulade-group-message-ids:{group_id}"

    def set_group_message_ids(self, group_id: str, message_ids: Iterable[str], ttl: int):
        key = self.build_group_message_id_key(group_id)
        return self._store([key], [message_ids], ttl)

    def get_group_message_ids(self, group_id: str):
        key = self.build_group_message_id_key(group_id)
        message_ids = self._get(key)
        if message_ids is Missing:
            raise MessageIdsMissing(f"Could't find message_ids for group {group_id}")
        return message_ids

    def delete_group_message_ids(self, group_id: str):
        key = self.build_group_message_id_key(group_id)
        self._delete(key)

    @staticmethod
    def build_group_completion_key(group_id: str) -> str:
        return f"remoulade-group-completion:{group_id}"

    def _get(self, message_key: str, forget: bool = False) -> Union[Type[Missing], Dict]:  # pragma: no cover
        """Get a result from the backend.  Subclasses may implement
        this method if they want to use the default, polling,
        implementation of get_result.
        """
        raise NotImplementedError(f"{type(self).__name__} does not implement _get()")

    def _store(self, message_keys: Iterable[str], result: Any, ttl: int) -> None:  # pragma: no cover
        """Store multiple results in the backend.  Subclasses may implement
        this method if they want to use the default implementation of
        set_result.
        """
        raise NotImplementedError(f"{type(self).__name__} does not implement _store()")

    def _delete(self, key: str) -> None:  # pragma: no cover
        """Delete a key from the backend"""
        raise NotImplementedError(f"{type(self).__name__} does not implement _delete()")

    @contextmanager
    def retry(self, broker, message, logger):
        try:
            yield
        except:  # noqa
            # If saving the result fail, we must retry the message else we have no way to know it's finished
            # We cannot use the retry middleware as it must be executed before the result middleware,
            # use a simplified one here
            retries = message.options.setdefault("retries_result_backend", 0)
            message.options["retries_result_backend"] = retries + 1
            if retries < 3:
                logger.error(f"Could not store result of {message}: retrying it", exc_info=True)
                _, backoff = compute_backoff(retries, min_backoff=500)  # retry after 500ms, 1s, 2s
                broker.enqueue(message, delay=backoff)
            else:
                logger.critical(f"Could not store result of {message}: retries exceeded", exc_info=True)



# This file is a part of Remoulade.
#
# Copyright (C) 2017,2018 CLEARTYPE SRL <bogdan@cleartype.io>
#
# Remoulade is free software; you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or (at
# your option) any later version.
#
# Remoulade is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public
# License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
from queue import Queue
from typing import TYPE_CHECKING, Dict, Iterable, List, Optional, Set, Type, TypeVar, Union

from .cancel import Cancel, CancelBackend
from .errors import ActorNotFound, NoCancelBackend, NoResultBackend, NoStateBackend
from .logging import get_logger
from .middleware import MiddlewareError, default_middleware
from .results import ResultBackend, Results
from .state import MessageState, StateBackend

if TYPE_CHECKING:
    from .actor import Actor
    from .message import Message  # noqa
    from .middleware import Middleware

    M = TypeVar("M", bound=Middleware)
#: The global broker instance.
global_broker: "Optional[Broker]" = None


def get_broker() -> "Broker":
    """Get the global broker instance.  If no global broker is set,
    this initializes a RabbitmqBroker and returns it.

    Returns:
      Broker: The default Broker.
    Raises:
      ValueError if no broker is set
    """
    global global_broker
    if global_broker is None:
        raise ValueError("Broker not found, are you sure you called set_broker(broker) ?")
    return global_broker


def set_broker(broker: "Broker") -> None:
    """Configure the global broker instance.

    Parameters:
      broker(Broker): The broker instance to use by default.
    """
    global global_broker
    global_broker = broker


def change_broker(broker: "Broker") -> None:
    """Change of broker instance

    It will take all the actors of the current broker and add them to the new one

    Parameters:
      broker(Broker): The broker instance to use by default.
    """
    global global_broker
    actors = global_broker.actors.values() if global_broker else []
    global_broker = broker
    declare_actors(actors)


def declare_actors(actors: "Iterable[Actor]") -> None:
    """Declare the given actors to the current broker

    Parameters:
      actors(List[Actor]): The actors being declared.
    """
    broker = get_broker()
    for actor in actors:
        broker.declare_actor(actor)


class Broker:
    """Base class for broker implementations.

    Parameters:
      middleware(list[Middleware]): The set of middleware that apply
        to this broker.  If you supply this parameter, you are
        expected to declare *all* middleware.  Most of the time,
        you'll want to use :meth:`.add_middleware` instead.

    Attributes:
      actor_options(set[str]): The names of all the options actors may
        overwrite when they are declared.
    """

    def __init__(self, middleware: "Optional[Iterable[Middleware]]" = None):
        self.logger = get_logger(__name__, type(self))
        self.actors: "Dict[str, Actor]" = {}
        self.queues: Dict[str, Optional[Queue]] = {}
        self.delay_queues: Set[str] = set()

        self.actor_options: Set[str] = set()
        self.middleware: "List[Middleware]" = []

        if middleware is None:
            middleware = [m() for m in default_middleware]

        for m in middleware:
            self.add_middleware(m)

    @property
    def local(self) -> bool:
        return False

    def emit_before(self, signal, *args, **kwargs):
        for middleware in self.middleware:
            try:
                getattr(middleware, "before_" + signal)(self, *args, **kwargs)
            except MiddlewareError:
                raise
            except Exception:
                self.logger.critical("Unexpected failure in before_%s.", signal, exc_info=True)

    def emit_after(self, signal, *args, **kwargs):
        for middleware in reversed(self.middleware):
            try:
                getattr(middleware, "after_" + signal)(self, *args, **kwargs)
            except Exception:
                self.logger.critical("Unexpected failure in after_%s.", signal, exc_info=True)

    def get_result_backend(self) -> ResultBackend:
        """Get the ResultBackend associated with the broker

        Raises:
            NoResultBackend: if there is no ResultBackend

        Returns:
            ResultBackend: the result backend
        """
        return self._get_backend("results")

    def get_cancel_backend(self) -> CancelBackend:
        """Get the CancelBackend associated with the broker

        Raises:
            NoCancelBackend: if there is no CancelBackend

        Returns:
            CancelBackend: the cancel backend
        """
        return self._get_backend("cancel")

    def get_state_backend(self) -> StateBackend:
        """Get the StateBackend associated with the broker

        Raises:
            NoStateBackend: if there is no StateBackend

        Returns:
            StateBackend: the state backend
        """
        return self._get_backend("state")

    def _get_backend(self, name: str):
        """Get the backend associated with the broker either cancel or results"""
        message = "The default broker doesn't have a %s backend."
        backends = {
            "results": (Results, NoResultBackend(message % "results")),
            "cancel": (Cancel, NoCancelBackend(message % "cancel")),
            "state": (MessageState, NoStateBackend(message % "state")),
        }
        try:
            middleware_class, exception = backends[name]
            middleware: Optional[Union[Results, Cancel, MessageState]] = self.get_middleware(middleware_class)
            if middleware is not None:
                return middleware.backend
            else:
                raise exception
        except KeyError:
            raise ValueError("invalid backend name")

    def add_middleware(
        self, middleware: "Middleware", *, before: "Type[Middleware]" = None, after: "Type[Middleware]" = None
    ) -> None:
        """Add a middleware object to this broker.  The middleware is
        appended to the end of the middleware list by default, or to
        the default point of the middleware.

        You can specify another middleware (by class) as a reference
        point for where the new middleware should be added.

        Parameters:
          middleware(Middleware): The middleware.
          before(type): Add this middleware before a specific one.
          after(type): Add this middleware after a specific one.

        Raises:
          ValueError: When either ``before`` or ``after`` refer to a
            middleware that hasn't been registered yet.
        """
        before = before or middleware.default_before
        after = after or middleware.default_after

        assert not (before and after), "provide either 'before' or 'after', but not both"

        if before or after:
            for i, m in enumerate(self.middleware):  # noqa
                if isinstance(m, before or after):
                    break
            else:
                raise ValueError("Middleware %r not found" % (before or after))

            if before:
                self.middleware.insert(i, middleware)
            else:
                self.middleware.insert(i + 1, middleware)
        else:
            self.middleware.append(middleware)

        self.actor_options |= middleware.actor_options

        for actor_name in self.get_declared_actors():
            middleware.after_declare_actor(self, actor_name)

        for queue_name in self.get_declared_queues():
            middleware.after_declare_queue(self, queue_name)

        for queue_name in self.get_declared_delay_queues():
            middleware.after_declare_delay_queue(self, queue_name)

    def get_middleware(self, middleware_class: "Type[M]") -> "Optional[M]":
        for middleware in self.middleware:
            if isinstance(middleware, middleware_class):
                return middleware
        return None

    def close(self):
        """Close this broker and perform any necessary cleanup actions."""

    def consume(self, queue_name: str, prefetch: int = 1, timeout: int = 30000) -> "Consumer":  # pragma: no cover
        """Get an iterator that consumes messages off of the queue.

        Raises:
          QueueNotFound: If the given queue was never declared.

        Parameters:
          queue_name(str): The name of the queue to consume messages off of.
          prefetch(int): The number of messages to prefetch per consumer.
          timeout(int): The amount of time in milliseconds to idle for.

        Returns:
          Consumer: A message iterator.
        """
        raise NotImplementedError

    def declare_actor(self, actor: "Actor") -> None:  # pragma: no cover
        """Declare a new actor on this broker.  Declaring an Actor
        twice replaces the first actor with the second by name.

        Parameters:
          actor(Actor): The actor being declared.
        """
        actor.set_broker(self)
        self.emit_before("declare_actor", actor)
        self.declare_queue(actor.queue_name)
        self.actors[actor.actor_name] = actor
        self.emit_after("declare_actor", actor)

    def declare_queue(self, queue_name: str) -> None:  # pragma: no cover
        """Declare a queue on this broker.  This method must be
        idempotent.

        Parameters:
          queue_name(str): The name of the queue being declared.
        """
        raise NotImplementedError

    def enqueue(self, message: "Message", *, delay: Optional[int] = None) -> "Message":  # pragma: no cover
        """Enqueue a message on this broker.

        Parameters:
          message(Message): The message to enqueue.
          delay(int): The number of milliseconds to delay the message for.

        Returns:
          Message: Either the original message or a copy of it.
        """
        raise NotImplementedError

    def get_actor(self, actor_name: str) -> "Actor":  # pragma: no cover
        """Look up an actor by its name.

        Parameters:
          actor_name(str): The name to look up.

        Raises:
          ActorNotFound: If the actor was never declared.

        Returns:
          Actor: The actor.
        """
        try:
            return self.actors[actor_name]
        except KeyError:
            raise ActorNotFound(actor_name) from None

    def get_declared_actors(self) -> Set[str]:  # pragma: no cover
        """Get all declared actors.

        Returns:
          set[str]: The names of all the actors declared so far on
          this Broker.
        """
        return set(self.actors.keys())

    def get_declared_queues(self) -> Set[str]:  # pragma: no cover
        """Get all declared queues.

        Returns:
          set[str]: The names of all the queues declared so far on
          this Broker.
        """
        return set(self.queues.keys())

    def get_declared_delay_queues(self) -> Set[str]:  # pragma: no cover
        """Get all declared delay queues.

        Returns:
          set[str]: The names of all the delay queues declared so far
          on this Broker.
        """
        return self.delay_queues.copy()

    def flush(self, queue_name: str) -> None:  # pragma: no cover
        """Drop all the messages from a queue.

        Parameters:
          queue_name(str): The name of the queue to flush.
        """
        raise NotImplementedError()

    def flush_all(self) -> None:  # pragma: no cover
        """Drop all messages from all declared queues."""
        raise NotImplementedError()

    def join(self, queue_name: str, *, timeout: Optional[int] = None) -> None:  # pragma: no cover
        """Wait for all the messages on the given queue to be processed.
        This method is only meant to be used in tests to wait for all the messages in a queue to be processed."""
        raise NotImplementedError()


class Consumer:
    """Consumers iterate over messages on a queue.

    Consumers and their MessageProxies are *not* thread-safe.
    """

    def __iter__(self):  # pragma: no cover
        """Returns this instance as a Message iterator."""
        return self

    def ack(self, message):  # pragma: no cover
        """Acknowledge that a message has been processed, removing it
        from the broker.

        Parameters:
          message(MessageProxy): The message to acknowledge.
        """
        raise NotImplementedError

    def nack(self, message):  # pragma: no cover
        """Move a message to the dead-letter queue.

        Parameters:
          message(MessageProxy): The message to reject.
        """
        raise NotImplementedError

    def requeue(self, messages):  # pragma: no cover
        """Move unacked messages back to their queues.  This is called
        by consumer threads when they fail or are shut down.  The
        default implementation does nothing.

        Parameters:
          messages(list[MessageProxy]): The messages to requeue.
        """

    def __next__(self):  # pragma: no cover
        """Retrieve the next message off of the queue.  This method
        blocks until a message becomes available.

        Returns:
          MessageProxy: A transparent proxy around a Message that can
          be used to acknowledge or reject it once it's done being
          processed.
        """
        raise NotImplementedError

    def close(self):
        """Close this consumer and perform any necessary cleanup actions."""


class MessageProxy:
    """Base class for messages returned by :meth:`Broker.consume`."""

    def __init__(self, message):
        self.failed = False
        self._message = message

    def fail(self):
        """Mark this message for rejection."""
        self.failed = True

    def __getattr__(self, name):
        return getattr(self._message, name)

    def __str__(self):
        return str(self._message)

    def __lt__(self, other):
        # This can get called if two messages have the same priority
        # in a queue.  If that's the case, we don't care which runs
        # first.
        return True

    def __eq__(self, other):
        if isinstance(other, MessageProxy):
            return self._message == other._message
        return self._message == other



import redis

from remoulade.common import chunk

from ..backend import State, StateBackend


class RedisBackend(StateBackend):
    """A state backend for Redis_.

    Parameters:
      namespace(str): A string with which to prefix result keys.
      encoder(Encoder): The encoder to use when storing and retrieving
        result data.  Defaults to :class:`.JSONEncoder`.
      client(Redis): An optional client.  If this is passed,
        then all other parameters are ignored.
      url(str): An optional connection URL.  If both a URL and
        connection parameters are provided, the URL is used.
      **parameters(dict): Connection parameters are passed directly
        to :class:`redis.Redis`.

    .. _redis: https://redis.io
    """

    def __init__(self, *, namespace="remoulade-state", encoder=None, client=None, url=None, **parameters):
        super().__init__(namespace=namespace, encoder=encoder)
        if url:
            parameters["connection_pool"] = redis.ConnectionPool.from_url(url)
        self.client = client or redis.Redis(**parameters)

    def get_state(self, message_id):
        key = self._build_message_key(message_id)
        data = self.client.hgetall(key)
        if not data:
            return None
        return self._parse_state(data)

    def set_state(self, state, ttl=3600):
        message_key = self._build_message_key(state.message_id)
        with self.client.pipeline() as pipe:
            encoded_state = self._encode_dict(state.as_dict())
            pipe.hset(message_key, mapping=encoded_state)
            pipe.expire(message_key, ttl)
            pipe.execute()

    def get_states(self):
        size = 1000
        for keys in chunk(self.client.scan_iter(match="{}*".format(StateBackend.namespace), count=size), size):
            with self.client.pipeline() as pipe:
                for key in keys:
                    pipe.hgetall(key)
                data = pipe.execute()
                for state in data:
                    if state:
                        yield self._parse_state(state)

    def _parse_state(self, data):
        decoded_state = self._decode_dict(data)
        return State.from_dict(decoded_state)



from marshmallow import Schema, ValidationError, fields, validate, validates_schema


class MessageSchema(Schema):
    """
    Class to validate post data in /messages
    """

    actor_name = fields.Str(validate=validate.Length(min=1), required=True)
    args = fields.List(fields.Raw(), allow_none=True)
    kwargs = fields.Dict(allow_none=True)
    options = fields.Dict(allow_none=True)
    delay = fields.Number(validate=validate.Range(min=1), allow_none=True)


class PageSchema(Schema):
    """
    Class to validate the attributes of a page
    """

    sort_column = fields.Str(allow_none=True)
    sort_direction = fields.Str(allow_none=True, validate=validate.OneOf(["asc", "desc"]))
    size = fields.Int(missing=50, validate=validate.Range(min=1, max=1000))
    search_value = fields.Str(allow_none=True)
    offset = fields.Int(missing=0)

    @validates_schema
    def validate_sort(self, data, **kwargs):
        if data.get("sort_direction") and data.get("sort_column") is None:
            raise ValidationError("sort_column is not defined")



from typing import Any, Dict, Set

from ..backend import ForgottenResult, Missing, ResultBackend


class LocalBackend(ResultBackend):
    """An in-memory result backend. For use with LocalBroker only.

    We need to be careful here: if an actor store its results and never forget it, we may store all its results
    and never delete it. Resulting in a memory leak.
    """

    results: Dict[str, Any] = {}
    group_completions: Dict[str, Set[str]] = {}
    forgotten_results: Set[str] = set()

    def _get(self, message_key: str, forget: bool = False):
        if message_key in self.forgotten_results:
            return ForgottenResult.asdict()

        try:
            if forget:
                data = self.results.pop(message_key)
                self.forgotten_results.add(message_key)
                return data
            else:
                return self.results[message_key]
        except KeyError:
            return Missing

    def _store(self, message_keys, results, _):
        for (message_key, result) in zip(message_keys, results):
            self.results[message_key] = result

    def _delete(self, key: str):
        del self.results[key]

    def increment_group_completion(self, group_id: str, message_id: str) -> int:
        group_completion_key = self.build_group_completion_key(group_id)
        completed = self.group_completions.get(group_completion_key, set()) | {message_id}
        self.group_completions[group_completion_key] = completed
        return len(completed)



# This file is a part of Remoulade.
#
# Copyright (C) 2017,2018 CLEARTYPE SRL <bogdan@cleartype.io>
#
# Remoulade is free software; you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or (at
# your option) any later version.
#
# Remoulade is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public
# License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
import os
import time
from threading import local

import prometheus_client as prom

from ..logging import get_logger
from .middleware import Middleware

#: The default HTTP host the exposition server should bind to.
DEFAULT_HTTP_HOST = os.getenv("remoulade_prom_host", "127.0.0.1")

#: The default HTTP port the exposition server should listen on.
DEFAULT_HTTP_PORT = int(os.getenv("remoulade_prom_port", "9191"))


class Prometheus(Middleware):
    """A middleware that exports stats via Prometheus_.

    Parameters:
      http_host(str): The host to bind the Prometheus exposition
        server on.  This parameter can also be configured via the
        ``remoulade_prom_host`` environment variable.
      http_port(int): The port on which the server should listen.
        This parameter can also be configured via the
        ``remoulade_prom_port`` environment variable.
      registry(CollectorRegistry): the prometheus registry to use, if None, use a new registry.

    .. _Prometheus: https://prometheus.io
    """

    def __init__(self, *, http_host=DEFAULT_HTTP_HOST, http_port=DEFAULT_HTTP_PORT, registry=None):
        self.logger = get_logger(__name__, type(self))
        self.http_host = http_host
        self.http_port = http_port
        self.local_data = local()
        self.registry = registry

        self.worker_busy = None
        self.total_errored_messages = None
        self.total_retried_messages = None
        self.total_rejected_messages = None
        self.message_durations = None

    @property
    def actor_options(self):
        """The set of options that may be configured on each actor."""
        return {"prometheus_label"}

    @staticmethod
    def _get_labels(broker, message):
        actor = broker.get_actor(message.actor_name)
        return message.queue_name, actor.options.get("prometheus_label", message.actor_name)

    def _init_labels(self, actor):
        # initialize the metrics for all labels to 0
        metrics = [
            self.total_errored_messages,
            self.total_retried_messages,
            self.total_rejected_messages,
            self.message_durations,
        ]
        for metric in metrics:
            if metric:  # metric can be None if actor is declared before worker boot
                metric.labels(actor.queue_name, actor.options.get("prometheus_label", actor.actor_name))

    def before_worker_boot(self, broker, worker):
        self.logger.debug("Setting up metrics...")
        if self.registry is None:
            self.registry = prom.CollectorRegistry()
        self.worker_busy = prom.Gauge(
            "remoulade_worker_busy", "1 if the worker is processing a message, 0 if not", registry=self.registry
        )
        self.total_errored_messages = prom.Counter(
            "remoulade_message_errors_total",
            "The total number of errored messages.",
            ["queue_name", "actor_name"],
            registry=self.registry,
        )
        self.total_retried_messages = prom.Counter(
            "remoulade_message_retries_total",
            "The total number of retried messages.",
            ["queue_name", "actor_name"],
            registry=self.registry,
        )
        self.total_rejected_messages = prom.Counter(
            "remoulade_message_rejects_total",
            "The total number of dead-lettered messages.",
            ["queue_name", "actor_name"],
            registry=self.registry,
        )
        self.message_durations = prom.Summary(
            "remoulade_message_duration_milliseconds",
            "The time spent processing messages.",
            ["queue_name", "actor_name"],
            registry=self.registry,
        )
        for actor in broker.actors.values():
            self._init_labels(actor)

        self.logger.debug("Starting exposition server...")
        prom.start_http_server(addr=self.http_host, port=self.http_port, registry=self.registry)

    def after_worker_boot(self, broker, worker):
        self.worker_busy.set(0)

    def after_worker_shutdown(self, broker, worker):
        self.worker_busy.set(0)
        self.logger.debug("Shutting down exposition server...")
        # Do not stop it actually

    def after_declare_actor(self, broker, actor):
        self._init_labels(actor)

    def after_nack(self, broker, message):
        labels = self._get_labels(broker, message)
        self.total_rejected_messages.labels(*labels).inc()

    def after_enqueue(self, broker, message, delay):
        if "retries" in message.options:
            labels = self._get_labels(broker, message)
            self.total_retried_messages.labels(*labels).inc()

    @property
    def message_start_times(self):
        if not hasattr(self.local_data, "message_start_times"):
            self.local_data.message_start_times = {}
        return self.local_data.message_start_times

    def before_process_message(self, broker, message):
        self.message_start_times[message.message_id] = time.monotonic() * 1000
        self.worker_busy.set(1)

    def after_process_message(self, broker, message, *, result=None, exception=None):
        labels = self._get_labels(broker, message)
        message_start_time = self.message_start_times.pop(message.message_id, None)
        message_duration = 0
        if message_start_time is not None:
            message_duration = (time.monotonic() * 1000) - message_start_time
        self.message_durations.labels(*labels).observe(message_duration)
        if exception is not None:
            self.total_errored_messages.labels(*labels).inc()
        if not self.message_start_times:
            self.worker_busy.set(0)

    after_skip_message = after_process_message
    after_message_canceled = after_process_message



# This file is a part of Remoulade.
#
# Copyright (C) 2017,2018 CLEARTYPE SRL <bogdan@cleartype.io>
#
# Remoulade is free software; you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or (at
# your option) any later version.
#
# Remoulade is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public
# License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
import time
from contextlib import contextmanager
from queue import Empty, Full, LifoQueue
from threading import Lock, local
from typing import TYPE_CHECKING, List, Optional

from amqpstorm import AMQPChannelError, AMQPConnectionError, AMQPError, UriConnection
from typing_extensions import Final

from ..broker import Broker, Consumer, MessageProxy
from ..common import current_millis, dq_name, xq_name
from ..errors import ChannelPoolTimeout, ConnectionClosed, QueueJoinTimeout
from ..logging import get_logger
from ..message import Message

if TYPE_CHECKING:
    from ..middleware import Middleware  # noqa

#: The maximum amount of time a message can be in the dead queue.
DEAD_MESSAGE_TTL: Final[int] = 86400000 * 7

#: The max number of times to attempt an enqueue operation in case of
#: a connection error.
MAX_ENQUEUE_ATTEMPTS: Final[int] = 6


class RabbitmqBroker(Broker):
    """A broker that can be used with RabbitMQ.

    Examples:

      >>> RabbitmqBroker(url="amqp://guest:guest@127.0.0.1:5672")

    Parameters:
      confirm_delivery(bool): Wait for RabbitMQ to confirm that
        messages have been committed on every call to enqueue.
        Defaults to False.
      url(str): The optional connection URL to use to determine which Rabbit server to connect to.
        If None is provided, connection is made with 'amqp://guest:guest@localhost:5672'
      middleware(list[Middleware]): The set of middleware that apply
        to this broker.
      max_priority(int): Configure the queues with x-max-priority to
        support priority queue in RabbitMQ itself
      channel_pool_size(int): Size of the channel pool
      dead_queue_max_length(int|None): Max size of the dead queue. If None, no max size.

    """

    def __init__(
        self,
        *,
        confirm_delivery: bool = False,
        url: Optional[str] = None,
        middleware: Optional[List["Middleware"]] = None,
        max_priority: Optional[int] = None,
        channel_pool_size: int = 200,
        dead_queue_max_length: Optional[int] = None,
    ):
        super().__init__(middleware=middleware)

        if max_priority is not None and not (0 < max_priority <= 255):
            raise ValueError("max_priority must be a value between 0 and 255")

        if dead_queue_max_length is not None and dead_queue_max_length <= 0:
            raise ValueError("dead_queue_max_length must be strictly above 0")

        self.url = url or ""
        self.confirm_delivery = confirm_delivery
        self.max_priority = max_priority
        self.dead_queue_max_length = dead_queue_max_length
        self._connection = None
        self.queues = {}
        self.state = local()
        self.channel_pool = ChannelPool(channel_factory=self.channel_factory, pool_size=channel_pool_size)
        self.queues_declared = False
        # we need a Lock on self._connection as it can be modified by multiple threads
        self.lock = Lock()

    @property
    def connection(self):
        """The :class:amqpstorm.Connection` for the current
        proccess.  This property may change without notice.
        """
        with self.lock:
            if self._connection is None or self._connection.is_closed:
                self._connection = UriConnection(self.url)
            return self._connection

    @connection.deleter
    def connection(self):
        with self.lock:
            if self._connection is not None:
                try:
                    self._connection.close()
                except AMQPError:
                    pass
            self._connection = None

    def channel_factory(self):
        channel = self.connection.channel()
        if self.confirm_delivery:
            channel.confirm_deliveries()
        return channel

    def close(self) -> None:
        """Close all open RabbitMQ connections."""

        self.logger.debug("Closing channels and connection...")
        try:
            del self.connection
        except Exception:  # pragma: no cover
            self.logger.debug("Encountered an error while connection.", exc_info=True)
        self._connection = None
        self.channel_pool.clear()
        self.logger.debug("Channels and connections closed.")

    def consume(self, queue_name: str, prefetch: int = 1, timeout: int = 5000) -> "_RabbitmqConsumer":
        """Create a new consumer for a queue.

        Parameters:
          queue_name(str): The queue to consume.
          prefetch(int): The number of messages to prefetch.
          timeout(int): The idle timeout in milliseconds.

        Returns:
          Consumer: A consumer that retrieves messages from RabbitMQ.
        """
        try:
            self._declare_rabbitmq_queues()
        except (AMQPConnectionError, AMQPChannelError) as e:
            if isinstance(e, AMQPConnectionError):
                del self.connection

            self.channel_pool.clear()

            raise ConnectionClosed(e) from None
        return _RabbitmqConsumer(self.connection, queue_name, prefetch, timeout)

    def _declare_rabbitmq_queues(self):
        """Real Queue declaration to happen before enqueuing or consuming

        Raises:
          AMQPConnectionError or AMQPChannelError: If the underlying channel or connection has been closed.
        """
        with self.channel_pool.acquire() as channel:
            for queue_name in self.queues:
                self._declare_queue(channel, queue_name)
                self._declare_dq_queue(channel, queue_name)
                self._declare_xq_queue(channel, queue_name)

    def declare_queue(self, queue_name: str) -> None:
        """Declare a queue.  Has no effect if a queue with the given
        name already exists.

        Parameters:
          queue_name(str): The name of the new queue.

        Raises:
          ConnectionClosed: If the underlying channel or connection
            has been closed.
        """
        if queue_name not in self.queues:
            self.emit_before("declare_queue", queue_name)
            self.queues[queue_name] = None
            self.emit_after("declare_queue", queue_name)

            delayed_name = dq_name(queue_name)
            self.delay_queues.add(delayed_name)
            self.emit_after("declare_delay_queue", delayed_name)

    def _build_queue_arguments(self, queue_name):
        arguments = {
            "x-dead-letter-exchange": "",
            "x-dead-letter-routing-key": xq_name(queue_name),
        }
        if self.max_priority:
            arguments["x-max-priority"] = self.max_priority

        return arguments

    def _declare_queue(self, channel, queue_name):
        arguments = self._build_queue_arguments(queue_name)
        return channel.queue.declare(queue=queue_name, durable=True, arguments=arguments)

    def _declare_dq_queue(self, channel, queue_name):
        arguments = self._build_queue_arguments(queue_name)
        return channel.queue.declare(queue=dq_name(queue_name), durable=True, arguments=arguments)

    def _declare_xq_queue(self, channel, queue_name):
        arguments = {
            # This HAS to be a static value since messages are expired
            # in order inside of RabbitMQ (head-first).
            "x-message-ttl": DEAD_MESSAGE_TTL,
        }
        if self.dead_queue_max_length:
            arguments["x-max-length"] = self.dead_queue_max_length
        return channel.queue.declare(queue=xq_name(queue_name), durable=True, arguments=arguments)

    def enqueue(self, message: "Message", *, delay: Optional[int] = None) -> "Message":
        """Enqueue a message.

        Parameters:
          message(Message): The message to enqueue.
          delay(int): The minimum amount of time, in milliseconds, to
            delay the message by.

        Raises:
          ConnectionClosed: If the underlying channel or connection
            has been closed.
        """
        queue_name = message.queue_name
        actor = self.get_actor(message.actor_name)
        properties = {"delivery_mode": 2, "priority": message.options.get("priority", actor.priority)}
        if delay is not None:
            queue_name = dq_name(queue_name)
            message_eta = current_millis() + delay
            message = message.copy(queue_name=queue_name, options={"eta": message_eta})

        attempts = 1
        while True:
            try:
                # I chose to do queue declaration only on first enqueuing, it should be sufficient but it do not
                # resolve the case of queue deletion at runtime. But we do not want the overhead of queue creation on
                # each enqueue
                if not self.queues_declared:
                    self._declare_rabbitmq_queues()
                    self.queues_declared = True
                self.logger.debug("Enqueueing message %r on queue %r.", message.message_id, queue_name)
                self.emit_before("enqueue", message, delay)
                with self.channel_pool.acquire() as channel:
                    channel.basic.publish(
                        exchange="", routing_key=queue_name, body=message.encode(), properties=properties
                    )
                self.emit_after("enqueue", message, delay)
                return message

            except (AMQPConnectionError, AMQPChannelError) as e:
                # Delete the channel (and the connection if needed) so that the
                # next caller/attempt may initiate new ones of each.
                if isinstance(e, AMQPConnectionError):
                    del self.connection

                self.channel_pool.clear()

                attempts += 1
                if attempts > MAX_ENQUEUE_ATTEMPTS:
                    raise ConnectionClosed(e) from None

                self.logger.debug("Retrying enqueue due to closed connection. [%d/%d]", attempts, MAX_ENQUEUE_ATTEMPTS)

    def get_queue_message_counts(self, queue_name: str):
        """Get the number of messages in a queue.  This method is only
        meant to be used in unit and integration tests.

        Parameters:
          queue_name(str): The queue whose message counts to get.

        Returns:
          tuple: A triple representing the number of messages in the
          queue, its delayed queue and its dead letter queue.
        """
        with self.channel_pool.acquire() as channel:
            queue_response = self._declare_queue(channel, queue_name)
            dq_queue_response = self._declare_dq_queue(channel, queue_name)
            xq_queue_response = self._declare_xq_queue(channel, queue_name)

        return (
            queue_response["message_count"],
            dq_queue_response["message_count"],
            xq_queue_response["message_count"],
        )

    def flush(self, queue_name: str) -> None:
        """Drop all the messages from a queue.

        Parameters:
          queue_name(str): The queue to flush.
        """
        for name in (queue_name, dq_name(queue_name), xq_name(queue_name)):
            with self.channel_pool.acquire() as channel:
                channel.queue.purge(name)

    def flush_all(self) -> None:
        """Drop all messages from all declared queues."""
        for queue_name in self.queues:
            self.flush(queue_name)

    def join(
        self, queue_name: str, min_successes: int = 10, idle_time: int = 100, *, timeout: Optional[int] = None
    ) -> None:
        """Wait for all the messages on the given queue to be
        processed.  This method is only meant to be used in tests to
        wait for all the messages in a queue to be processed.

        Warning:
          This method doesn't wait for unacked messages so it may not
          be completely reliable.  Use the stub broker in your unit
          tests and only use this for simple integration tests.

        Parameters:
          queue_name(str): The queue to wait on.
          min_successes(int): The minimum number of times all the
            polled queues should be empty.
          idle_time(int): The number of milliseconds to wait between
            counts.
          timeout(Optional[int]): The max amount of time, in
            milliseconds, to wait on this queue.
        """
        deadline = timeout and time.monotonic() + timeout / 1000
        successes = 0
        while successes < min_successes:
            if deadline and time.monotonic() >= deadline:
                raise QueueJoinTimeout(queue_name)

            total_messages = sum(self.get_queue_message_counts(queue_name)[:-1])
            if total_messages == 0:
                successes += 1
            else:
                successes = 0

            time.sleep(idle_time / 1000)


class _RabbitmqConsumer(Consumer):
    def __init__(self, connection, queue_name, prefetch, timeout):
        try:
            self.logger = get_logger(__name__, type(self))
            self.channel = connection.channel()
            self.channel.basic.qos(prefetch_count=prefetch)
            self.channel.basic.consume(queue=queue_name, no_ack=False)
            self.timeout = timeout
        except (AMQPConnectionError, AMQPChannelError) as e:
            raise ConnectionClosed(e) from None

    def ack(self, message):
        try:
            message.ack()
        except (AMQPConnectionError, AMQPChannelError) as e:
            raise ConnectionClosed(e) from None
        except Exception:  # pragma: no cover
            self.logger.warning("Failed to ack message.", exc_info=True)

    def nack(self, message):
        try:
            message.nack(requeue=False)
        except (AMQPConnectionError, AMQPChannelError) as e:
            raise ConnectionClosed(e) from None
        except Exception:  # pragma: no cover
            self.logger.warning("Failed to nack message.", exc_info=True)

    def requeue(self, messages):
        """RabbitMQ automatically re-enqueues unacked messages when
        consumers disconnect so this is a no-op.
        """

    def __next__(self):
        """Return None if no value after timeout seconds"""
        try:
            deadline = time.monotonic() + self.timeout / 1000
            message = None
            while message is None and time.monotonic() < deadline:
                try:
                    message = next(self.channel.build_inbound_messages(auto_decode=False, break_on_empty=True))
                except StopIteration:
                    time.sleep(0.1)

            return _RabbitmqMessage(message) if message else None
        except (AMQPConnectionError, AMQPChannelError) as e:
            raise ConnectionClosed(e) from None

    def close(self):
        try:
            self.channel.close()
        except (AMQPConnectionError, AMQPChannelError):
            pass


class _RabbitmqMessage(MessageProxy):
    def __init__(self, rabbitmq_message):
        super().__init__(Message.decode(rabbitmq_message.body))

        self._rabbitmq_message = rabbitmq_message

    def ack(self):
        self._rabbitmq_message.ack()

    def nack(self, requeue):
        self._rabbitmq_message.nack(requeue)


class ChannelPool:
    """A pool of channels that can be used by the RabbitmqBroker.

    The pool uses a synchronized queue as a backend, making sure that two threads never end up sharing a channel.

    The channels are created lazily as the reservation requests comes.

    The ChannelPool should be used via the `acquire` context manager, to make sure that a used channel is properly put
    back into the pool.

    Examples:

      >>> channel_pool = ChannelPool(channel_factory=lambda connection: connection.channel(), pool_size=5)
      >>> with channel_pool.acquire() as channel:
      ...     channel.basic.publish(...)

    Parameters:
      channel_factory(function): Function that will be called to create new channels.
      pool_size(int): The max size of the pool.

    """

    def __init__(self, channel_factory, *, pool_size):
        self._channel_factory = channel_factory
        self._pool = LifoQueue(pool_size)  # type: LifoQueue
        self._pool_size = pool_size

        for _ in range(pool_size):
            # The goal is to create lazily the channels, so Nones are put as stand-in for the channel that can be used
            self.put(None)

    @contextmanager
    def acquire(self, timeout=None):
        """
        Parameters:
          timeout(int): The max number of second to wait when fetching a channel from the pool. If None, it will wait
            indefinitely to get a channel. Default None.

        Raises:
          ChannelPoolTimeout: when the timeout for reserving a channel is run out.
        """
        channel = self.get(timeout=timeout)
        if channel is None:
            channel = self._channel_factory()
        try:
            yield channel
        finally:
            if channel.is_closed:
                self.put(None)
            else:
                self.put(channel)

    def get(self, *, timeout=None):
        try:
            return self._pool.get(timeout=timeout)
        except Empty:
            raise ChannelPoolTimeout("Could not get any channel from the pool")

    def put(self, channel):
        try:
            return self._pool.put_nowait(channel)
        except Full:
            pass

    def __len__(self):
        return self._pool.qsize()

    def clear(self):
        """
        This will empty the pool and fill it back with None.

        It is best to use it inside a lock to avoid doing it multiple times.
        """
        while len(self) > 0:
            channel = self._pool.get_nowait()
            if channel is not None:
                try:
                    channel.close()
                except AMQPError:
                    pass

        for _ in range(self._pool_size):
            self.put(None)



# This file is a part of Remoulade.
#
# Copyright (C) 2017,2018 CLEARTYPE SRL <bogdan@cleartype.io>
#
# Remoulade is free software; you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or (at
# your option) any later version.
#
# Remoulade is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public
# License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import traceback
from typing import Callable

from ..helpers import compute_backoff
from ..logging import get_logger
from .middleware import Middleware

#: The default minimum amount of backoff to apply to retried tasks.
DEFAULT_MIN_BACKOFF = 15000

#: The default maximum amount of backoff to apply to retried tasks.
DEFAULT_MAX_BACKOFF = 1000 * 60 * 60


class Retries(Middleware):
    """Middleware that automatically retries failed tasks with
    exponential backoff.

    Parameters:
      max_retries(int): The maximum number of times tasks can be retried.
      min_backoff(int): The minimum amount of backoff milliseconds to
        apply to retried tasks.  Defaults to 15 seconds.
      max_backoff(int): The maximum amount of backoff milliseconds to
        apply to retried tasks.  Defaults to 7 days.
      retry_when(Callable[[int, Exception], bool]): An optional
        predicate that can be used to programmatically determine
        whether a task should be retried or not.  This takes
        precedence over `max_retries` when set.
    """

    def __init__(
        self,
        *,
        max_retries: int = 0,
        min_backoff: int = None,
        max_backoff: int = None,
        retry_when: Callable[[int, Exception], bool] = None,
        backoff_strategy: str = "exponential",
        jitter: bool = True,
    ):
        self.logger = get_logger(__name__, type(self))
        self.max_retries = max_retries
        self.min_backoff = min_backoff or DEFAULT_MIN_BACKOFF
        self.max_backoff = max_backoff or DEFAULT_MAX_BACKOFF
        self.retry_when = retry_when
        self.backoff_strategy = backoff_strategy
        self.jitter = jitter

    @property
    def actor_options(self):
        return {"max_retries", "min_backoff", "max_backoff", "retry_when", "backoff_strategy", "jitter"}

    def after_process_message(self, broker, message, *, result=None, exception=None):
        if exception is None:
            return

        retries = message.options.setdefault("retries", 0)
        max_retries = self.get_option("max_retries", broker=broker, message=message)
        retry_when = self.get_option("retry_when", broker=broker, message=message)
        if (
            retry_when is not None
            and not retry_when(retries, exception)
            or retry_when is None
            and max_retries is not None
            and retries >= max_retries
        ):
            if max_retries > 0:
                self.logger.warning("Retries exceeded for message %r.", message.message_id)
            message.fail()
            return

        message.options["retries"] += 1
        message.options["traceback"] = traceback.format_exc(limit=30)
        min_backoff = self.get_option("min_backoff", broker=broker, message=message)
        max_backoff = self.get_option("max_backoff", broker=broker, message=message)
        backoff_strategy = self.get_option("backoff_strategy", broker=broker, message=message)
        jitter = self.get_option("jitter", broker=broker, message=message)
        max_retries = self.get_option("max_retries", broker=broker, message=message)
        _, backoff = compute_backoff(
            retries,
            min_backoff=min_backoff,
            max_backoff=max_backoff,
            jitter=jitter,
            max_retries=max_retries,
            backoff_strategy=backoff_strategy,
        )
        self.logger.info("Retrying message %r in %d milliseconds.", message.message_id, backoff)
        broker.enqueue(message, delay=backoff)



import datetime
import json
import time
from typing import Dict, List, Union

import pytz
import redis

from remoulade import Broker, get_encoder, get_logger

DEFAULT_JOB_INTERVAL = 3600 * 24
DEFAULT_JOB_STATUS = True
DEFAULT_TZ = "UTC"

DEFAULT_SCHEDULER_NAMESPACE = "remoulade-schedule"
DEFAULT_SCHEDULER_LOCK_KEY = "remoulade-scheduler-tick"
DEFAULT_SCHEDULER_PERIOD = 1


class ScheduledJob:
    def __init__(
        self,
        actor_name: str,
        args: List = None,
        kwargs: Dict = None,
        interval: int = None,
        daily_time: datetime.time = None,
        iso_weekday: int = None,
        enabled: bool = None,
        last_queued: datetime.datetime = None,
        tz: str = None,
    ) -> None:
        """
        Describes a job that should be run through the scheduler
        :param actor_name: The name of the actor that should be ran
        :param args: args that should be passed
        :param kwargs: kwargs that should be passed
        :param interval: at which interval (in seconds) this should run
        :param daily_time: at which time of day this should run (this assumes interval=24H)
        :param iso_weekday: the isoweekday this should be running at
        :param enabled: allows to disable jobs
        :param last_queued: when is the last time this job was run through the scheduler ?
        you should not be setting this yourself
        :param tz: timezone your daily_time is expressed in
        """
        if (daily_time is not None and daily_time.tzinfo is not None) or (
            last_queued is not None and last_queued.tzinfo is not None
        ):
            raise ValueError("Datetimes should be naive")
        if daily_time is not None and interval not in (3600 * 24, None):
            raise ValueError("daily_time should be used with a 24h interval")
        self.actor_name = actor_name
        self.interval = interval if interval is not None else DEFAULT_JOB_INTERVAL
        self.daily_time = daily_time
        self.iso_weekday = iso_weekday
        self.enabled = enabled if enabled is not None else DEFAULT_JOB_STATUS
        self.last_queued = last_queued
        self.tz = tz if tz is not None else DEFAULT_TZ
        self.args = args if args is not None else []
        self.kwargs = kwargs if kwargs is not None else {}

    def get_hash(self) -> str:
        d = self.as_dict()
        args = json.dumps(d["args"])
        kwargs = json.dumps(sorted(list(d["kwargs"].items()), key=lambda x: x[0]))

        path = [str(d[k]) for k in ("actor_name", "interval", "daily_time", "iso_weekday", "enabled", "tz")] + [
            args,
            kwargs,
        ]

        return " ".join(path)

    def as_dict(self) -> Dict:
        return {
            "actor_name": self.actor_name,
            "interval": self.interval,
            "daily_time": None if self.daily_time is None else self.daily_time.strftime("%H:%M:%S"),
            "iso_weekday": self.iso_weekday,
            "enabled": self.enabled,
            "last_queued": None if self.last_queued is None else self.last_queued.strftime("%Y-%m-%dT%H:%M:%S"),
            "tz": self.tz,
            "args": self.args,
            "kwargs": self.kwargs,
        }

    def encode(self) -> bytes:
        return get_encoder().encode(self.as_dict())

    @classmethod
    def decode(cls, data: bytes) -> "ScheduledJob":
        data = get_encoder().decode(data)
        return ScheduledJob(
            actor_name=data["actor_name"],
            interval=data["interval"],
            daily_time=None
            if data["daily_time"] is None
            else datetime.datetime.strptime(data["daily_time"], "%H:%M:%S").time(),
            iso_weekday=data["iso_weekday"],
            enabled=data["enabled"],
            last_queued=None
            if data["last_queued"] is None
            else datetime.datetime.strptime(data["last_queued"], "%Y-%m-%dT%H:%M:%S"),
            tz=data["tz"],
            args=data["args"],
            kwargs=data["kwargs"],
        )


class Scheduler:
    def __init__(
        self,
        broker: Broker,
        schedule: List[ScheduledJob],
        *,
        namespace: str = None,
        lock_key: str = None,
        period: Union[int, float] = None,
        client: redis.Redis = None,
        url: str = None,
        **redis_parameters,
    ) -> None:
        """
        The class that takes care of scheduling tasks
        :param broker: The broker for the remoulade app
        :param schedule: A map that gives the schedule you want to be running
        :param namespace: Where we should store our scheduled tasks on redis
        :param lock_key: The redis key that we use as a lock to make sure only one scheduler is working at a time
        :param period: At which period (in seconds) the scheduler should be running, default is 1 second
        :param client: The redis client we should use to access redis
        :param url: The URL of the redis client
        :param redis_parameters: Additional params to pass to redis
        """
        self.schedule = schedule
        self.namespace = namespace if namespace is not None else DEFAULT_SCHEDULER_NAMESPACE
        self.period = period if period is not None else DEFAULT_SCHEDULER_PERIOD
        self.broker = broker
        self.lock_key = lock_key if lock_key is not None else DEFAULT_SCHEDULER_LOCK_KEY

        if url:
            redis_parameters["connection_pool"] = redis.ConnectionPool.from_url(url)

        self.client = client or redis.Redis(**redis_parameters)
        self.logger = get_logger(__name__, type(self))

    def flush(self, job: ScheduledJob) -> None:
        self.client.hset(self.namespace, job.get_hash().encode("utf-8"), job.encode())

    def get_redis_schedule(self) -> Dict[str, ScheduledJob]:
        r = {}
        for json_schedule in self.client.hgetall(self.namespace).values():
            job = ScheduledJob.decode(json_schedule)
            r[job.get_hash()] = job
        return r

    def sync_config(self) -> None:
        redis_schedule = self.get_redis_schedule()
        config_schedule = {job.get_hash(): job for job in self.schedule}

        # loop over redis schedule
        for job_hash in redis_schedule.keys():
            # if job is no longer in configured schedule, remove it
            if job_hash not in config_schedule:
                self.logger.info("Dropping %s because it's no longer in the config", job_hash)
                self.client.hdel(self.namespace, job_hash.encode("utf-8"))

        # add newly configured jobs
        for job_hash, job in config_schedule.items():
            if job_hash not in redis_schedule:
                # Do not queue task if daily time is already passed
                if job.daily_time is not None and job.daily_time < datetime.datetime.now(pytz.timezone(job.tz)).time():
                    self.logger.info(
                        "Will not run %s today, because daily time has already passed. Wait for tomorrow", job_hash
                    )
                    job.last_queued = datetime.datetime.utcnow()
                # Add to redis
                self.logger.info("Adding new job %s to schedule", job_hash)
                self.flush(job)

    def start(self):
        self.logger.debug("Syncing config")
        self.sync_config()

        while True:
            with self.client.lock(self.lock_key, timeout=10, blocking_timeout=20):
                schedule = self.get_redis_schedule()
                for job_hash, job in schedule.items():
                    now = datetime.datetime.now(pytz.timezone(job.tz))
                    now_utc = datetime.datetime.utcnow()
                    curr_isodow = now.isoweekday()

                    if not job.enabled:
                        continue

                    # Tasks that are supposed to run at a specific time/day
                    if job.iso_weekday is not None or job.daily_time is not None:
                        if job.iso_weekday is not None and job.iso_weekday != curr_isodow:
                            continue

                        if job.daily_time is not None and now.time() < job.daily_time:
                            continue

                        if job.last_queued is not None:
                            # if task already ran today, skip it
                            last_queued_date = (
                                job.last_queued.replace(tzinfo=pytz.UTC).astimezone(pytz.timezone(job.tz)).date()
                            )
                            if now.date() == last_queued_date:
                                continue
                    # Task that should run each X seconds
                    else:
                        if job.last_queued is not None and (
                            datetime.timedelta(seconds=0)
                            < now_utc - job.last_queued
                            < datetime.timedelta(seconds=job.interval)
                        ):
                            continue

                    if job.actor_name not in self.broker.actors:
                        self.logger.error(f"Unknown {job_hash}. Cannot queue it.")
                        continue

                    self.broker.actors[job.actor_name].send(*job.args, **job.kwargs)

                    job.last_queued = datetime.datetime.utcnow()
                    self.flush(job)

            time.sleep(self.period)



# This file is a part of Remoulade.
#
# Copyright (C) 2017,2018 CLEARTYPE SRL <bogdan@cleartype.io>
#
# Remoulade is free software; you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or (at
# your option) any later version.
#
# Remoulade is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public
# License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
from ..errors import NoResultBackend
from ..logging import get_logger
from .middleware import Middleware


class Pipelines(Middleware):
    """Middleware that lets you pipe actors together so that the
    output of one actor feeds into the input of another.

    Parameters:
      pipe_ignore(bool): When True, ignores the result of the previous
        actor in the pipeline.
      pipe_target(dict): A message representing the actor the current
        result should be fed into.
    """

    def __init__(self):
        self.logger = get_logger(__name__, type(self))

    @property
    def actor_options(self):
        return {
            "pipe_ignore",
            "pipe_target",
        }

    def after_process_message(self, broker, message, *, result=None, exception=None):
        from ..composition import GroupInfo

        if exception is not None or getattr(message, "failed", False):
            return

        pipe_target = self.get_option("pipe_target", broker=broker, message=message)
        group_info = self.get_option("group_info", broker=broker, message=message)
        if group_info:
            group_info = GroupInfo(**group_info)

        if pipe_target is not None:

            try:
                if group_info and not self._group_completed(message, group_info, broker):
                    return

                self._send_next_message(pipe_target, broker, result, group_info)

                broker.emit_after("enqueue_pipe_target", group_info)

            except NoResultBackend as e:
                self.logger.error(str(e))
                message.fail()

    def _send_next_message(self, pipe_target, broker, result, group_info):
        """Send a message to the pipe target (if it's a list to all the pipe targets)

        If it's a group, we need to get the result of each actor from the result backend because it's not present in
        the message.

        Parameters:
            pipe_target(dict|List[dict]): the message or list of message to enqueue
            broker(Broker): the broker
            result(any): the result of the actor (to be send with the message if pipe_ignore=False)
            group_info(GroupInfo|None): the info of the group if it's a group else None
        """
        # Since Pipelines is a default middleware, this import has to
        # happen at runtime in order to avoid a cyclic dependency
        # from broker -> pipelines -> messages -> broker.
        from ..message import Message

        for message_data in pipe_target:
            next_message = Message(**message_data)
            pipe_ignore = self.get_option("pipe_ignore", broker=broker, message=next_message, default=False)

            if not pipe_ignore:
                if group_info:
                    result = self._group_results(group_info, broker)
                next_message = next_message.copy(args=next_message.args + (result,))

            broker.enqueue(next_message)

    @staticmethod
    def _group_results(group_info, broker):
        """Get the result of a group (fetch the group members message_ids then all the results)"""
        from ..collection_results import CollectionResults

        try:
            result_backend = broker.get_result_backend()
        except NoResultBackend:
            raise NoResultBackend("Pipeline with groups are ony supported with a result backend")

        message_ids = result_backend.get_group_message_ids(group_id=group_info.group_id)
        results = CollectionResults.from_message_ids(message_ids)
        return list(results.get())

    def _group_completed(self, message, group_info, broker):
        """Returns true if a group is completed, and increment the completion count of the group

        Parameters:
            group_info(GroupInfo): the info of the group to get the completion from
            broker(Broker): the broker to use

        Raises:
            NoResultBackend: if there is no result backend set
        """
        try:
            result_backend = broker.get_result_backend()
        except NoResultBackend:
            raise NoResultBackend("Pipeline with groups are ony supported with a result backend")

        with result_backend.retry(broker, message, self.logger):
            group_completion = result_backend.increment_group_completion(group_info.group_id, message.message_id)
        group_competed = group_completion >= group_info.children_count
        if group_competed:
            self.logger.info("Finished group %s.", group_info.group_id)
        if group_completion == 1:
            self.logger.info("First message of Group %s has been completed.", group_info.group_id)
        return group_competed



from random import uniform

from typing_extensions import Literal

from remoulade.errors import UnknownStrategy

BackoffStrategy = Literal["constant", "linear", "spread_linear", "exponential", "spread_exponential"]


def compute_backoff(
    attempts: int,
    *,
    backoff_strategy: BackoffStrategy = "exponential",
    min_backoff: int = 5,
    max_backoff: int = 2000,
    max_retries: int = 32,
    jitter: bool = True,
):
    """Compute the backoff using the selected strategy.

    Parameters:
      attempts(int): The number of attempts there have been so far.
      backoff_strategy(str): The selected strategy.
      min_backoff(int): The minimum backoff duration in milliseconds.
      jitter(bool): If true adds a small random value to the backoff to avoid mass simultaneous retries.
      max_backoff(int): The max number of milliseconds to backoff by.
      max_retries(int): The maximum number of retries.
    """

    if backoff_strategy == "spread_exponential":
        backoff = compute_backoff_spread_exponential(
            attempts, min_backoff=min_backoff, max_backoff=max_backoff, max_retries=max_retries
        )
    elif backoff_strategy == "constant":
        backoff = min_backoff
    elif backoff_strategy == "linear":
        backoff = min(attempts * min_backoff, max_backoff)
    elif backoff_strategy == "spread_linear":
        backoff = compute_backoff_spread_linear(
            attempts, min_backoff=min_backoff, max_backoff=max_backoff, max_retries=max_retries
        )
    elif backoff_strategy == "exponential":
        backoff = compute_backoff_exponential(
            attempts, min_backoff=min_backoff, max_backoff=max_backoff, max_retries=max_retries
        )
    else:
        raise UnknownStrategy(f"Unknown retry strategy: {backoff_strategy}")
    if jitter:
        jitter_factor = 1 / 4
        backoff = int(backoff * (1 - jitter_factor) + uniform(0, backoff) * jitter_factor)

    return attempts + 1, backoff


def compute_backoff_spread_linear(
    attempts: int, *, min_backoff: int = 5, max_backoff: int = 2000, max_retries: int = 32
):
    """Compute an linear backoff value linearly spread between min_backoff and max_backoff.

    Parameters:
      attempts(int): The number of attempts there have been so far.
      min_backoff(int): The minimum backoff duration in milliseconds.
      max_backoff(int): The max number of milliseconds to backoff by.
      max_retries(int): The maximum number of retries.

    Returns:
      int: The backoff in milliseconds.
    """

    if max_retries == 0:
        return min(attempts * min_backoff, max_backoff)
    if max_retries == 1:
        return min_backoff
    else:
        return min_backoff + (max_backoff - min_backoff) * min((attempts - 1) / (max_retries - 1), 1)


def compute_backoff_exponential(attempts: int, *, min_backoff: int = 5, max_backoff: int = 2000, max_retries: int = 32):
    """Compute an exponential backoff value based on some number of attempts.

    Parameters:
      attempts(int): The number of attempts there have been so far.
      min_backoff(int): The minimum backoff duration in milliseconds.
      max_backoff(int): The max number of milliseconds to backoff by.
      max_retries(int): The maximum number of retries.

    Returns:
      int: The backoff in milliseconds.
    """
    exponent = min(attempts - 1, max_retries - 1)
    backoff = min(min_backoff * 2 ** exponent, max_backoff)
    return backoff


def compute_backoff_spread_exponential(
    attempts: int, *, min_backoff: int = 5, max_backoff: int = 2000, max_retries: int = 32
):
    """Compute an exponential backoff value exponentially spread between min_backoff and max_backoff.

    Parameters:
      attempts(int): The number of attempts there have been so far.
      min_backoff(int): The minimum backoff duration in milliseconds.
      max_backoff(int): The max number of milliseconds to backoff by.
      max_retries(int): The maximum number of retries.

    Returns:
      int: The backoff in milliseconds.
    """

    if max_retries == 1:
        return min_backoff
    exponent = min((attempts - 1) / (max_retries - 1), 1)
    backoff = min_backoff * ((max_backoff / min_backoff) ** exponent)
    return backoff



# This file is a part of Remoulade.
#
# Copyright (C) 2017,2018 CLEARTYPE SRL <bogdan@cleartype.io>
#
# Remoulade is free software; you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or (at
# your option) any later version.
#
# Remoulade is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public
# License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import resource

from ..logging import get_logger
from .middleware import Middleware


class MaxMemory(Middleware):
    """Middleware that stop a worker if its amount of resident memory exceed max_memory (in kilobytes)
    If a task causes a worker to exceed this limit, the task will be completed, and the worker will be
    stopped afterwards.

    Parameters:
      max_memory(int): The maximum amount of resident memory (in kilobytes)
    """

    def __init__(self, *, max_memory: int):
        self.logger = get_logger(__name__, type(self))
        self.max_memory = max_memory

    def after_worker_thread_process_message(self, broker, thread):
        used_memory = resource.getrusage(resource.RUSAGE_SELF).ru_maxrss
        if used_memory <= 0:
            self.logger.error("Worker unable to determine memory usage")
        if used_memory > self.max_memory:
            self.logger.warning(
                f"Stopping worker thread as used_memory ({used_memory:.2E}Kb) > max_memory ({self.max_memory:.2E}Kb)"
            )
            # stopping the worker thread will in time stop the worker process which check if all workers are still
            # running (via Worker.worker_stopped)
            thread.stop()



# This file is a part of Remoulade.
#
# Copyright (C) 2017,2018 WIREMIND SAS <dev@wiremind.fr>
#
# Remoulade is free software; you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or (at
# your option) any later version.
#
# Remoulade is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public
# License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
import time
from typing import Dict, Iterable, Optional

from ..backend import CancelBackend


class StubBackend(CancelBackend):
    """An in-memory cancel backend.  For use in unit tests.

    Parameters:
      cancellation_ttl(int): The minimal amount of seconds message cancellations
        should be kept in the backend.
    """

    def __init__(self, *, cancellation_ttl: Optional[int] = None) -> None:
        super().__init__(cancellation_ttl=cancellation_ttl)
        self.cancellations: Dict[str, float] = {}

    def is_canceled(self, message_id: str, group_id: str) -> bool:
        return any(
            self.cancellations.get(key, -float("inf")) > time.time() - self.cancellation_ttl
            for key in [message_id, group_id]
            if key
        )

    def cancel(self, message_ids: Iterable[str]) -> None:
        timestamp = time.time()
        for message_id in message_ids:
            self.cancellations[message_id] = timestamp



# This file is a part of Remoulade.
#
# Copyright (C) 2017,2018 WIREMIND SAS <dev@wiremind.fr>
#
# Remoulade is free software; you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or (at
# your option) any later version.
#
# Remoulade is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public
# License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
from typing import Iterable, Optional

DEFAULT_CANCELLATION_TTL = 3600


class CancelBackend:
    """ABC for cancel backends.

    Parameters:
      cancellation_ttl(int): The minimal amount of seconds message cancellations
        should be kept in the backend (default 1h).
    """

    def __init__(self, *, cancellation_ttl: Optional[int] = None) -> None:
        self.cancellation_ttl = cancellation_ttl or DEFAULT_CANCELLATION_TTL

    def is_canceled(self, message_id: str, group_id: str) -> bool:
        """Return true if the message has been canceled"""
        raise NotImplementedError("%(classname)r does not implement is_canceled" % {"classname": type(self).__name__})

    def cancel(self, message_ids: Iterable[str]) -> None:
        """Mark a message as canceled"""
        raise NotImplementedError("%(classname)r does not implement cancel" % {"classname": type(self).__name__})



# This file is a part of Remoulade.
#
# Copyright (C) 2017,2018 CLEARTYPE SRL <bogdan@cleartype.io>
#
# Remoulade is free software; you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or (at
# your option) any later version.
#
# Remoulade is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public
# License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


class RemouladeError(Exception):  # pragma: no cover
    """Base class for all remoulade errors."""

    def __init__(self, message: str) -> None:
        self.message = message

    def __str__(self) -> str:
        return str(self.message) or repr(self.message)


class BrokerError(RemouladeError):
    """Base class for broker-related errors."""


class ActorNotFound(BrokerError):
    """Raised when a message is sent to an actor that hasn't been declared."""


class QueueNotFound(BrokerError):
    """Raised when a message is sent to an queue that hasn't been declared."""


class QueueJoinTimeout(RemouladeError):
    """Raised by brokers that support joining on queues when the join
    operation times out.
    """


class UnknownStrategy(RemouladeError):
    """Raised when the backoff_strategy option is set to an unexpected value."""


class ConnectionError(BrokerError):
    """Base class for broker connection-related errors."""


class ConnectionFailed(ConnectionError):
    """Raised when a broker connection could not be opened."""


class ConnectionClosed(ConnectionError):
    """Raised when a broker connection is suddenly closed."""


class RateLimitExceeded(RemouladeError):
    """Raised when a rate limit has been exceeded."""


class NoResultBackend(BrokerError):
    """Raised when trying to access the result backend on a broker without it"""


class NoCancelBackend(BrokerError):
    """Raised when trying to access the cancel backend on a broker without it"""


class NoStateBackend(BrokerError):
    """Raised when trying to access the state backend on a broker without it"""


class ChannelPoolTimeout(BrokerError):
    """Raised when the broker has wait for tool long to fetch a channel from its channel pool."""


class InvalidProgress(RemouladeError):
    """Raised when trying to set a progress that is greater than 1 or less than 0"""


class NoScheduler(RemouladeError):
    """Raised when trying to get a scheduler when there is not it"""
